from itertools import groupby
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np

from . import time_step
from .core.expressions import STOP, MathExpression
from .core.parser import ExpressionParser
from .core.rule import BaseRule, ExpressionChangeRule
from .rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
    VariableMultiplyRule,
)
from .state import MathyEnvState, MathyEnvStateStep, MathyObservation
from .types import EnvRewards, MathyEnvProblem, MathyEnvProblemArgs
from .util import (
    compare_expression_string_values,
    is_terminal_transition,
    print_error,
    raise_with_history,
)


class MathyEnv:
    """Implement a math solving game where a player wins by executing the
    right sequence of actions to reduce a math expression to an agreeable
    basic representation in as few moves as possible."""

    rules: List[BaseRule]
    max_moves: int
    verbose: bool
    reward_discount: float
    parser: ExpressionParser
    valid_actions_mask_cache: Dict[str, List[int]]
    valid_rules_cache: Dict[str, List[int]]

    def __init__(
        self,
        rules: List[BaseRule] = None,
        max_moves: int = 20,
        verbose: bool = False,
        error_invalid: bool = False,
        reward_discount: float = 0.99,
    ):
        self.discount = reward_discount
        self.verbose = verbose
        self.max_moves = max_moves
        self.error_invalid = error_invalid
        self.parser = ExpressionParser()
        if rules is None:
            self.rules = MathyEnv.core_rules()
        else:
            self.rules = rules
        self.valid_actions_mask_cache = dict()
        self.valid_rules_cache = dict()

    @classmethod
    def core_rules(cls, preferred_term_commute: bool = False) -> List[BaseRule]:
        """Return the mathy core agent actions"""
        return [
            ConstantsSimplifyRule(),
            CommutativeSwapRule(preferred=preferred_term_commute),
            DistributiveMultiplyRule(),
            DistributiveFactorOutRule(),
            AssociativeSwapRule(),
            VariableMultiplyRule(),
        ]

    @property
    def action_size(self) -> int:
        """Return the number of available actions"""
        return len(self.rules)

    def finalize_state(self, state: MathyEnvState):
        """Perform final checks on a problem state, to ensure the episode yielded
        results that were uncorrupted by transformation errors."""
        from_timestep: MathyEnvStateStep = state.agent.history[0]
        to_timestep: MathyEnvStateStep = state.agent.history[-1]
        compare_expression_string_values(
            str(from_timestep.raw), str(to_timestep.raw), state.agent.history
        )

    def get_env_namespace(self) -> str:
        """Return a unique dot namespaced string representing the current
        environment. e.g. mycompany.envs.differentiate"""
        raise NotImplementedError("subclass must implement this")

    def get_rewarding_actions(self, state: MathyEnvState) -> List[Type[BaseRule]]:
        """Get the list of rewarding action types. When these actions
        are selected, the agent gets a positive reward. """
        # NOTE: by default we give a positive reward for most actions taken. Reward
        #       values are only applied AFTER penalties, so things like reentrant
        #       states become negative reward even if their action is otherwise
        #       rewarding.
        return [
            ConstantsSimplifyRule,
            DistributiveFactorOutRule,
            VariableMultiplyRule,
        ]

    def get_penalizing_actions(self, state: MathyEnvState) -> List[Type[BaseRule]]:
        """Get the list of penalizing action types. When these actions
        are selected, the agent gets a negative reward."""
        return [
            AssociativeSwapRule,
            DistributiveMultiplyRule,
        ]

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        """Return the environment specific maximum move count for a given prolem."""
        return problem.complexity * 3

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """Provide environment-specific transitions per timestep."""
        return None

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Return a problem for the environment given a set of parameters
        to control problem generation.

        This is implemented per environment so each environment can
        generate its own dataset with no required configuration."""
        raise NotImplementedError("This must be implemented in a subclass")

    def state_to_observation(self, state: MathyEnvState,) -> MathyObservation:
        """Convert an environment state into an observation that can be used
        by a training agent."""

        action_mask = self.get_valid_moves(state)
        observation = state.to_observation(move_mask=action_mask, parser=self.parser)
        return observation

    def get_win_signal(self, env_state: MathyEnvState) -> float:
        """Calculate the reward value for completing the episode. This is done
        so that the reward signal can be scaled based on the time it took to
        complete the episode. """
        tiny = 3e-10
        total_moves = max(tiny, env_state.max_moves)
        # guard against divide by zero with max and a small value
        current_move = max(tiny, total_moves - env_state.agent.moves_remaining)
        bonus = (total_moves / current_move) / total_moves
        # If the episode is not very short, and the agent completes in half
        # the number of allowed steps, double the bonus signal
        if total_moves > 10 and current_move < total_moves / 2:
            bonus *= 2
        return min(2.0, EnvRewards.WIN + bonus)

    def get_lose_signal(self, env_state: MathyEnvState) -> float:
        """Calculate the reward value for failing to complete the episode. This is done
        so that the reward signal can be problem-type dependent."""
        return EnvRewards.LOSE

    def get_state_transition(
        self, env_state: MathyEnvState, searching: bool = False
    ) -> time_step.TimeStep:
        """Given an input state calculate the transition value of the timestep.

        # Parameters
            env_state: current env_state
            searching: True when called by MCTS simulation

        # Returns
            transition: the current state value transition
        """
        agent = env_state.agent
        expression = self.parser.parse(agent.problem)
        features = env_state.to_observation(
            self.get_valid_moves(env_state), parser=self.parser
        )
        root = expression.get_root()

        # Subclass specific win conditions happen here. Custom win-conditions
        # outside of that can override this method entirely.
        result = self.transition_fn(env_state, root, features)
        if result is not None:
            return result

        # Check the turn count last because if the previous move that incremented
        # the turn over the count resulted in a win-condition, we want it to be honored.
        if env_state.agent.moves_remaining <= 0:
            return time_step.termination(features, self.get_lose_signal(env_state))

        # The agent is penalized for returning to a previous state.
        for key, group in groupby(
            sorted([f"{h.raw}" for h in env_state.agent.history])
        ):
            list_count = len(list(group))
            if list_count <= 1 or key != expression.raw:
                continue

            # NOTE: the reward is scaled by how many times this state has been visited
            return time_step.transition(
                features,
                reward=EnvRewards.PREVIOUS_LOCATION * list_count,
                discount=self.discount,
            )

        if len(agent.history) > 0:
            last_timestep = agent.history[-1]
            rule = self.get_rule_from_timestep(last_timestep)
            reward_actions = self.get_rewarding_actions(env_state)
            # The rewarding_actions can be user specified
            for rewarding_class in reward_actions:
                if isinstance(rule, rewarding_class):
                    return time_step.transition(
                        features,
                        reward=EnvRewards.HELPFUL_MOVE,
                        discount=self.discount,
                    )

            penalty_actions = self.get_penalizing_actions(env_state)
            # The rewarding_actions can be user specified
            for penalty_class in penalty_actions:
                if isinstance(rule, penalty_class):
                    return time_step.transition(
                        features,
                        reward=EnvRewards.UNHELPFUL_MOVE,
                        discount=self.discount,
                    )

        # We're in a new state, and the agent is a little older.
        return time_step.transition(
            features, reward=EnvRewards.TIMESTEP, discount=self.discount
        )

    def get_next_state(
        self, env_state: MathyEnvState, action: int, searching: bool = False
    ) -> Tuple[MathyEnvState, time_step.TimeStep, ExpressionChangeRule]:
        """
        # Parameters
        env_state: current env_state
        action:    action taken
        searching: boolean set to True when called by MCTS

        # Returns
        next_state: env_state after applying action

        transition: the timestep that represents the state transition
        
        change: the change descriptor describing the change that happened
        """
        agent = env_state.agent
        expression = self.parser.parse(agent.problem)
        action_index, token_index = self.get_action_indices(action)
        token = self.get_token_at_index(expression, token_index)
        operation = self.rules[action_index]

        op_not_rule = not isinstance(operation, BaseRule)
        op_cannot_apply = token is None or operation.can_apply_to(token) is False
        if token is None or op_not_rule or op_cannot_apply:
            if self.error_invalid:
                steps = int(env_state.max_moves - agent.moves_remaining)
                msg = "Step: {} - Invalid action({}) '{}' for expression '{}'.".format(
                    steps, action, type(operation), expression
                )
                raise_with_history("Invalid Action", msg, agent.history)
                raise ValueError(f"Invalid Action: {msg}")
            else:
                valid_mask = self.get_valid_moves(env_state)
                # Non-masked searches ignore invalid moves entirely
                out_env = MathyEnvState.copy(env_state)
                out_env.action = -1
                obs = out_env.to_observation(
                    self.get_valid_moves(out_env), parser=self.parser
                )
                transition = time_step.transition(obs, EnvRewards.INVALID_MOVE)
                return out_env, transition, ExpressionChangeRule(BaseRule())

        change = operation.apply_to(token.clone_from_root())
        root = change.result.get_root()
        change_name = operation.name
        out_problem = str(root)
        out_env = env_state.get_out_state(
            problem=out_problem,
            focus=token_index,
            action=action_index,
            moves_remaining=agent.moves_remaining - 1,
        )

        transition = self.get_state_transition(out_env, searching)
        if not searching and self.verbose:
            token_idx = int("{}".format(token_index).zfill(3))
            self.print_state(
                out_env, change_name[:25].lower(), token_idx, change, transition.reward
            )
        return out_env, transition, change

    def print_state(
        self,
        env_state: MathyEnvState,
        action_name: str,
        token_index: int = -1,
        change: ExpressionChangeRule = None,
        change_reward: float = 0.0,
    ):
        """Render the given state to stdout for visualization"""
        print(
            self.render_state(
                env_state, action_name, token_index, change, change_reward
            )
        )

    def is_terminal_state(self, env_state: MathyEnvState) -> bool:
        """Determine if a given state is terminal or not.

        # Arguments
        env_state (MathyEnvState): The state to inspect

        # Returns
        (bool): A boolean indicating if the state is terminal or not.
        """
        return is_terminal_transition(self.get_state_transition(env_state))

    def print_history(self, env_state: MathyEnvState) -> None:
        """Render the history of an episode from a given state.

        # Arguments
        env_state (MathyEnvState): The state to render the history of.
        """
        history: List[MathyEnvStateStep] = env_state.agent.history[:]
        initial_step: MathyEnvStateStep = history.pop(0)
        curr_state: MathyEnvState = MathyEnvState(
            problem=initial_step.raw, max_moves=env_state.max_moves,
        )
        self.print_state(curr_state, "initial-state")
        while not self.is_terminal_state(curr_state):
            step: MathyEnvStateStep = history.pop(0)
            curr_state, transition, change = self.get_next_state(
                curr_state, step.action + (step.focus * len(self.rules))
            )
            rule_idx, token_idx = self.get_action_indices(step.action)
            rule: BaseRule = self.rules[rule_idx]
            self.print_state(
                env_state=curr_state,
                action_name=rule.name[:25].lower(),
                token_index=int(f"{step.focus}".zfill(3)),
                change=change,
                change_reward=transition.reward,
            )

    def render_state(
        self,
        env_state: MathyEnvState,
        action_name: str,
        token_index: int = -1,
        change: ExpressionChangeRule = None,
        change_reward: float = 0.0,
    ) -> str:
        """Render the given state to a string suitable for printing to a log"""
        changed_problem = env_state.agent.problem
        if change is not None and change.result is not None:
            changed_problem = change.result.get_root().terminal_text
        output = """{:<25} | {}""".format(action_name.lower(), changed_problem)

        def get_move_shortname(index, move):
            if move == 0:
                return "--"
            if move >= len(self.rules):
                return "xx"
            return self.rules[index].code.lower()

        token = "{}".format(token_index).zfill(3)
        moves_left = str(env_state.agent.moves_remaining).zfill(2)
        valid_rules = self.get_valid_rules(env_state)
        valid_moves = self.get_valid_moves(env_state)
        num_moves = "{}".format(len(np.nonzero(valid_moves)[0])).zfill(3)
        move_codes = [get_move_shortname(i, m) for i, m in enumerate(valid_rules)]
        moves = " ".join(move_codes)
        reward = f"{change_reward:.2}"
        reward = f"{reward:<5}"
        return f"{num_moves} | {moves} | {moves_left} | {token} | {reward} | {output}"

    def random_action(self, expression: MathExpression, rule: Type[BaseRule]) -> int:
        """Get a random action index that represents a particular rule"""

        found = False
        for r in self.rules:
            if isinstance(r, rule):
                found = True
                break
        if found is False:
            raise ValueError(
                "The action {rule} does not exist in the environment rule list"
            )
        actions = np.nonzero(self.get_actions_for_node(expression, [rule]))
        action = np.random.choice(actions[0])
        return action

    def get_initial_state(
        self, params: Optional[MathyEnvProblemArgs] = None, print_problem: bool = True
    ) -> Tuple[MathyEnvState, MathyEnvProblem]:
        """Generate an initial MathyEnvState for an episode"""
        config = params if params is not None else MathyEnvProblemArgs()
        prob: MathyEnvProblem = self.problem_fn(config)
        self.valid_actions_mask_cache = dict()
        self.valid_rules_cache = dict()
        self.parser.clear_cache()
        self.max_moves = self.max_moves_fn(prob, config)

        # Build and return the initial state
        env_state = MathyEnvState(
            problem=prob.text,
            problem_type=self.get_env_namespace(),
            max_moves=self.max_moves,
            num_rules=len(self.rules),
        )
        if print_problem and self.verbose:
            self.print_state(env_state, "initial-state")
        return env_state, prob

    def get_agent_actions_count(self, env_state: MathyEnvState) -> int:
        """Return number of all possible actions"""
        node_count = len(self.parser.parse(env_state.agent.problem).to_list())
        return self.action_size * node_count

    def get_token_at_index(
        self, expression: MathExpression, index: int
    ) -> Optional[MathExpression]:
        """Get the token that is `index` from the left of the expression"""
        count = 0
        result = None

        def visit_fn(node, depth, data):
            nonlocal result, count
            result = node
            if count == index:
                return STOP
            count = count + 1

        expression.visit_inorder(visit_fn)
        return result

    def get_valid_moves(self, env_state: MathyEnvState) -> List[int]:
        """Get a vector the length of the action space that is filled
         with 1/0 indicating whether the action at that index is valid
         for the current state.
        """
        agent = env_state.agent
        expression = self.parser.parse(agent.problem)
        return self.get_actions_for_node(expression)

    def get_valid_rules(self, env_state: MathyEnvState) -> List[int]:
        """Get a vector the length of the number of valid rules that is
        filled with 0/1 based on whether the rule has any nodes in the
        expression that it can be applied to.

        !!! note

            If you want to get a list of which nodes each rule can be
            applied to, prefer to use the `get_valid_moves` method.
        """
        key = self.to_hash_key(env_state)
        if key in self.valid_rules_cache:
            return self.valid_rules_cache[key]
        expression = self.parser.parse(env_state.agent.problem)
        actions = [0] * len(self.rules)
        for rule_index, rule in enumerate(self.rules):
            nodes = rule.find_nodes(expression)
            actions[rule_index] = 0 if len(nodes) == 0 else 1
        self.valid_rules_cache[key] = actions[:]
        return actions

    def get_action_indices(self, action: int) -> Tuple[int, int]:
        """Get the normalized action/node_index values from a
        given absolute action value.

        Returns a tuple of (rule_index, node_index)"""
        rule_count = len(self.rules)
        # Rule index = val % rule_count
        action_index = action % rule_count
        # And the action at that token
        token_index = int((action - action_index) / rule_count)
        return action_index, token_index

    def get_rule_from_timestep(self, time_step: MathyEnvStateStep) -> BaseRule:
        return self.rules[time_step.action]

    def get_actions_for_node(
        self, expression: MathExpression, rule_list: List[Type[BaseRule]] = None,
    ) -> List[int]:
        """Return a valid actions mask for the given expression and rule list. 

        Action masks are 1d lists of length (nodes * num_rules) where a 0 indicates
        the action is not valid in the current state, and a 1 indicates that it is
        a valid action to take."""
        key = str(expression)
        if rule_list is None and key in self.valid_actions_mask_cache:
            return self.valid_actions_mask_cache[key][:]
        node_count = len(expression.to_list())
        rule_count = len(self.rules)
        actions = [0] * rule_count * node_count
        for rule_index, rule in enumerate(self.rules):
            if rule_list is not None and not isinstance(rule, tuple(rule_list)):
                continue
            nodes = rule.find_nodes(expression)
            for node in nodes:
                action_index = (node.r_index * rule_count) + rule_index
                actions[action_index] = 1
        if rule_list is None:
            self.valid_actions_mask_cache[key] = actions[:]
        return actions

    def to_hash_key(self, env_state: MathyEnvState) -> str:
        """Convert env_state to a string for MCTS cache"""
        return env_state.agent.problem



import shutil
import tempfile
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from ..mathy.cli import cli
from ..mathy.envs.gym import MathyGymEnv  # noqa


def test_cli_contribute():
    runner = CliRunner()

    with patch("webbrowser.open") as mock:
        result = runner.invoke(cli, ["contribute"])
        mock.assert_called_with("https://mathy.ai/contributing", new=2)
        assert result.exit_code == 0


def test_cli_problems():
    runner = CliRunner()
    for problem_type in ["poly", "poly-combine", "complex", "binomial"]:
        result = runner.invoke(cli, ["problems", problem_type, "--number=100"])
        assert result.exit_code == 0


def test_cli_simplify():
    runner = CliRunner()
    for problem in ["4x + 2x", "(4 + 2) * x"]:
        result = runner.invoke(cli, ["simplify", problem, "--max-steps=3"])
        assert result.exit_code == 0


@pytest.mark.parametrize("agent", ["a3c", "zero"])
def test_cli_train(agent: str):
    runner = CliRunner()
    model_folder = tempfile.mkdtemp()
    result = runner.invoke(
        cli,
        [
            "train",
            agent,
            "poly-like-terms-haystack,poly-grouping",
            model_folder,
            "--verbose",
            "--mcts-sims=3",
            "--episodes=2",
            "--self-play-problems=1",
            "--training-iterations=1",
            "--workers=1",
        ],
    )
    assert result.exit_code == 0

    # Comment this out to keep your model
    shutil.rmtree(model_folder)



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utilities for Reinforcement Learning ops."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports

from six.moves import zip
import tensorflow as tf

LossOutput = collections.namedtuple("loss_output", ["loss", "extra"])


def best_effort_shape(tensor, with_rank=None):
    """Extract as much static shape information from a tensor as possible.

  Args:
    tensor: A `Tensor`. If `with_rank` is None, must have statically-known
        number of dimensions.
    with_rank: Optional, an integer number of dimensions to force the shape to
        be. Useful for tensors with no static shape information that must be
        of a particular rank. Default is None (number of dimensions must be
        statically known).

  Returns:
    An iterable with length equal to the number of dimensions in `tensor`,
    containing integers for the dimensions with statically-known size, and
    scalar `Tensor`s for dimensions with size only known at run-time.

  Raises:
    ValueError: If `with_rank` is None and `tensor` does not have
      statically-known number of dimensions.
  """
    tensor_shape = tensor.get_shape()
    if with_rank:
        tensor_shape = tensor_shape.with_rank(with_rank)
    if tensor_shape.ndims is None:
        raise ValueError(
            "`tensor` does not have statically-known number of dimensions."
        )
    shape_list = tensor_shape.as_list()
    for idx, dim in enumerate(shape_list):
        if not dim:
            shape_list[idx] = tf.shape(input=tensor)[idx]
    return shape_list


def assert_rank_and_shape_compatibility(tensors, rank):
    """Asserts that the tensors have the correct rank and compatible shapes.

  Shapes (of equal rank) are compatible if corresponding dimensions are all
  equal or unspecified. E.g. `[2, 3]` is compatible with all of `[2, 3]`,
  `[None, 3]`, `[2, None]` and `[None, None]`.

  Args:
    tensors: List of tensors.
    rank: A scalar specifying the rank that the tensors passed need to have.

  Raises:
    ValueError: If the list of tensors is empty or fail the rank and mutual
      compatibility asserts.
  """
    if not tensors:
        raise ValueError("List of tensors should be non-empty.")

    union_of_shapes = tf.TensorShape(None)
    for tensor in tensors:
        tensor_shape = tensor.get_shape()
        tensor_shape.assert_has_rank(rank)
        union_of_shapes = union_of_shapes.merge_with(tensor_shape)


def wrap_rank_shape_assert(tensors_list, expected_ranks, op_name):
    try:
        for tensors, rank in zip(tensors_list, expected_ranks):
            assert_rank_and_shape_compatibility(tensors, rank)
    except ValueError as e:
        error_message = "{}: Error in rank and/or " "compatibility check, {}".format(
            op_name, e
        )
        tf.compat.v1.logging.error(error_message)
        raise ValueError(error_message)


def assert_arg_bounded(value, min_value, max_value, op_name, arg_name):
    if not min_value <= value <= max_value:
        raise ValueError(
            (
                op_name
                + ": "
                + arg_name
                + " has to lie in "
                + "["
                + str(min_value)
                + ", "
                + str(max_value)
                + "]."
            )
        )



from typing import List, Optional

from numpy.random import randint, uniform

from .. import time_step
from ..core.expressions import MathExpression
from ..env import MathyEnv, MathyEnvProblem
from ..problems import gen_simplify_multiple_terms
from ..state import MathyEnvState, MathyObservation
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from ..util import get_terms, has_like_terms, is_preferred_term_form


class PolySimplify(MathyEnv):
    """A Mathy environment for simplifying polynomial expressions.

    NOTE: This environment only generates polynomial problems with
     addition operations. Subtraction, Multiplication and Division
     operators are excluded. This is a good area for improvement.
    """

    def __init__(self, ops: List[str] = None, **kwargs):
        super().__init__(**kwargs)
        if ops is None:
            ops = ["+"]
        self.ops = ops

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.simplify"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity * 6

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If there are no like terms."""
        if not has_like_terms(expression):
            term_nodes = get_terms(expression)
            is_win = True
            for term in term_nodes:
                if not is_preferred_term_form(term):
                    is_win = False
            if is_win:
                return time_step.termination(features, self.get_win_signal(env_state))
        return None

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Given a set of parameters to control term generation, produce
        a polynomial problem with (n) total terms divided among (m) groups
        of like terms. A few examples of the form: `f(n, m) = p`
        - (3, 1) = "4x + 2x + 6x"
        - (6, 4) = "4x + v^3 + y + 5z + 12v^3 + x"
        - (4, 2) = "3x^3 + 2z + 12x^3 + 7z"
        """
        noise = uniform(0.85, 1.0)

        if params.difficulty == MathyEnvDifficulty.easy:
            noise_terms = randint(1, 3)
            num_terms = randint(3, 6)
            scaling = uniform(0.35, 0.5)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op=self.ops,
                inner_terms_scaling=scaling,
                powers_probability=0.2,
                noise_probability=0.5,
                shuffle_probability=0.0,
                noise_terms=noise_terms,
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            # Set number of noise terms and inject a bunch more in simple problems
            # so they don't learn stupid policies that only work with small trees.
            #
            # e.g. mashing the 3rd node to commute the tree until a DF shows up in
            #      the desired position.
            noise_terms = randint(1, 4)
            num_terms = randint(4, 8)
            scaling = uniform(0.5, 0.6)
            powers = uniform(0.35, 0.8)
            shuffle = uniform(0.35, 0.8)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op=self.ops,
                inner_terms_scaling=scaling,
                powers_probability=powers,
                noise_probability=noise,
                shuffle_probability=shuffle,
                noise_terms=noise_terms,
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            noise_terms = randint(3, 7)
            num_terms = randint(6, 10)
            scaling = uniform(0.5, 0.5)
            powers = uniform(0.15, 0.8)
            shuffle = uniform(0.35, 0.9)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op=self.ops,
                inner_terms_scaling=scaling,
                powers_probability=powers,
                noise_probability=noise,
                shuffle_probability=shuffle,
                noise_terms=noise_terms,
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, complexity, self.get_env_namespace())



from ..mathy import MathyEnvState


def test_env_state():
    state = MathyEnvState(problem="4+4")
    assert state is not None


def test_env_state_encode_player():
    env_state = MathyEnvState(problem="4x+2")
    env_state = env_state.get_out_state(
        problem="2+4x", focus=0, moves_remaining=10, action=0
    )
    agent = env_state.agent
    assert agent.problem == "2+4x"
    assert agent.moves_remaining == 10
    assert agent.action == 0


def test_env_state_serialize_string():
    env_state = MathyEnvState(problem="4x+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="2+4x", focus=i, moves_remaining=10 - i, action=i
        )

    state_str = env_state.to_string()
    compare = MathyEnvState.from_string(state_str)
    assert env_state.agent.problem == compare.agent.problem
    assert env_state.agent.moves_remaining == compare.agent.moves_remaining
    for one, two in zip(env_state.agent.history, compare.agent.history):
        assert one.raw == two.raw
        assert one.focus == two.focus
        assert one.action == two.action


def test_env_state_serialize_numpy():
    env_state = MathyEnvState(problem="4x+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="2+4x", focus=i, moves_remaining=10 - i, action=i
        )

    state_np = env_state.to_np()
    compare = MathyEnvState.from_np(state_np)
    assert env_state.agent.problem == compare.agent.problem
    assert env_state.agent.moves_remaining == compare.agent.moves_remaining
    for one, two in zip(env_state.agent.history, compare.agent.history):
        assert one.raw == two.raw
        assert one.focus == two.focus
        assert one.action == two.action


def test_env_state_to_observation():
    """to_observation has defaults to allow calling with no arguments"""
    env_state = MathyEnvState(problem="4x+2")
    assert env_state.to_observation() is not None



from multiprocessing import Queue
from typing import Dict, List, Optional

import numpy as np

from ..state import (
    MathyObservation,
    MathyWindowObservation,
    observations_to_window,
)


class EpisodeMemory(object):
    # Observation from the environment
    observations: List[MathyObservation]
    # Action taken in the environment
    actions: List[int]
    # Reward from the environmnet
    rewards: List[float]
    # Estimated value from the model
    values: List[float]

    def __init__(self):
        self.clear()

    def clear(self):
        self.observations = []
        self.actions = []
        self.rewards = []
        self.values = []

    def clear_except_window(self, window_size: int):
        """Clear all data except for a window's worth of elements"""
        self.observations = self.observations[-window_size:]
        self.actions = self.actions[-window_size:]
        self.rewards = self.rewards[-window_size:]
        self.values = self.values[-window_size:]

    def to_window_observation(
        self, observation: MathyObservation, window_size: int = 3
    ) -> MathyWindowObservation:
        previous = -(max(window_size - 1, 1))
        window_observations = self.observations[previous:] + [observation]
        assert len(window_observations) <= window_size
        return observations_to_window(window_observations)

    def to_episode_window(self) -> MathyWindowObservation:
        return observations_to_window(self.observations)

    def store(
        self,
        *,
        observation: MathyObservation,
        action: int,
        reward: float,
        value: float,
    ):
        self.observations.append(observation)
        self.actions.append(action)
        self.rewards.append(reward)
        self.values.append(value)



import math
from typing import List, Optional, Type, Union

import gym
import numpy as np
from gym.envs.registration import register

from ...core.expressions import MathExpression
from ...core.rule import ExpressionChangeRule
from ...env import MathyEnv
from ...state import MathyEnvState, MathyObservation, observations_to_window
from ...types import MathyEnvProblem, MathyEnvProblemArgs
from ...util import discount, is_terminal_transition, pad_array
from .masked_discrete import MaskedDiscrete


class MathyGymEnv(gym.Env):
    """A small wrapper around Mathy envs to allow them to work with OpenAI Gym. The
    agents currently use this env wrapper, but it could be dropped in the future. """

    mathy: MathyEnv
    state: Optional[MathyEnvState]
    _challenge: Optional[MathyEnvState]
    env_class: Type[MathyEnv]
    env_problem_args: Optional[MathyEnvProblemArgs]

    def __init__(
        self,
        env_class: Type[MathyEnv] = MathyEnv,
        env_problem_args: Optional[MathyEnvProblemArgs] = None,
        np_observation: bool = False,
        repeat_problem: bool = False,
        **env_kwargs,
    ):
        self.state = None
        self.repeat_problem = repeat_problem
        self.np_observation = np_observation
        self.mathy = env_class(**env_kwargs)
        self.env_class = env_class
        self.env_problem_args = env_problem_args
        self._challenge, _ = self.mathy.get_initial_state(env_problem_args)
        self.action_space = MaskedDiscrete(self.action_size, [1] * self.action_size)

    @property
    def action_size(self) -> int:
        if self.state is not None:
            return self.mathy.get_agent_actions_count(self.state)
        return self.mathy.action_size

    def step(self, action: int):
        assert self.state is not None, "call reset() before stepping the environment"
        self.state, transition, change = self.mathy.get_next_state(self.state, action)
        done = is_terminal_transition(transition)
        info = {
            "transition": transition,
            "done": done,
            "valid": change.result is not None,
        }
        if done:
            info["win"] = transition.reward > 0.0
        return self._observe(self.state), transition.reward, done, info

    def _observe(self, state: MathyEnvState) -> Union[MathyObservation, np.ndarray]:
        """Observe the environment at the given state, updating the observation
        space and action space for the given state. """
        action_mask = self.mathy.get_valid_moves(state)
        observation = self.mathy.state_to_observation(state)
        self.action_space.n = self.mathy.get_agent_actions_count(state)
        self.action_space.mask = action_mask
        if self.np_observation:
            # convert mask to probabilities
            nodes = np.array(pad_array(observation.nodes, 512, 0))
            mask = np.array(pad_array(observation.mask, 512, 0))
            mask = mask / np.sum(mask)
            return np.vstack((nodes, mask))
        return observation

    def reset(self):
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        if self.repeat_problem:
            self.state = MathyEnvState.copy(self._challenge)
        else:
            self.state, self.problem = self.mathy.get_initial_state(
                self.env_problem_args
            )
        return self._observe(self.state)

    def reset_with_input(self, problem_text: str, max_moves=16):
        # If the episode is being reset because it ended, assert the validity
        # of the last problem outcome
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        self.reset()
        self.state = MathyEnvState(problem=problem_text, max_moves=max_moves)
        return self._observe(self.state)

    def render(
        self,
        last_action: int = -1,
        last_reward: float = 0.0,
        last_change: Optional[ExpressionChangeRule] = None,
    ):
        assert self.state is not None, "call reset() before rendering the env"
        action_name = "initial"
        token_index = -1
        if last_action != -1:
            action_index, token_index = self.mathy.get_action_indices(last_action)
            action_name = self.mathy.rules[action_index].name
        else:
            print(f"Problem: {self.state.agent.problem}")
        self.mathy.print_state(
            self.state,
            action_name[:25].lower(),
            token_index,
            change=last_change,
            change_reward=last_reward,
        )


def safe_register(id: str, **kwargs):
    """Ignore re-register errors."""
    try:
        return register(id, **kwargs)
    except gym.error.Error:
        pass



import sys
from typing import NamedTuple
from enum import Enum
from pydantic import BaseModel, Field

# Use typing_extensions for Python < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Final, Literal
else:
    from typing_extensions import Final, Literal  # noqa


class MathyEnvDifficulty(str, Enum):
    easy: str = Field(
        default="easy",
        title="Easy Problems",
        description="The simplest form of problems that demonstrate a task",
    )
    normal: str = Field(
        default="normal",
        title="Normal Problems",
        description="Challenging problems that involve more terms",
    )
    hard: str = Field(
        default="hard",
        title="Hard Problems",
        description="Difficult problems that have intentionally large expressions",
    )


class MathyEnvProblemArgs(BaseModel):
    difficulty: MathyEnvDifficulty = MathyEnvDifficulty.easy


class MathyEnvProblem(NamedTuple):
    """Summarize an environment-specific problem that was generated with
    a tuple of (text, complexity, type) where:
     - "text" is the text content of the generated problem
     - "complexity" is an integer value that represents the number of
       terms in the problem text.
     - "type" is a dot namespaced string, e.g. "mathy.poly.simplify"
    """

    text: str
    complexity: int
    type: str


class EnvRewards:
    LOSE = -1.0
    WIN = 1.0
    HELPFUL_MOVE = 0.01
    UNHELPFUL_MOVE = -0.01
    TIMESTEP = -0.01
    PREVIOUS_LOCATION = -0.02
    INVALID_MOVE = -0.5



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tensorflow ops for common Distributional RL learning rules.

Distributions are taken to be categorical over a support of 'N' distinct atoms,
which are always specified in ascending order.

These ops define state/action value distribution learning rules for discrete,
scalar, action spaces. Actions must be represented as indices in the range
`[0, K)` where `K` is the number of distinct actions.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
import tensorflow as tf
from . import base_ops
from . import distribution_ops

Extra = collections.namedtuple("dist_value_extra", ["target"])

_l2_project = distribution_ops.l2_project


def _slice_with_actions(embeddings, actions):
    """Slice a Tensor.

  Take embeddings of the form [batch_size, num_actions, embed_dim]
  and actions of the form [batch_size, 1], and return the sliced embeddings
  like embeddings[:, actions, :].

  Args:
    embeddings: Tensor of embeddings to index.
    actions: int Tensor to use as index into embeddings

  Returns:
    Tensor of embeddings indexed by actions
  """
    shape = tuple(t.value for t in embeddings.get_shape())
    batch_size, num_actions = shape[0], shape[1]

    # Values are the 'values' in a sparse tensor we will be setting
    act_indx = tf.cast(actions, tf.int64)[:, None]
    values = tf.reshape(tf.cast(tf.ones(tf.shape(input=actions)), tf.bool), [-1])

    # Create a range for each index into the batch
    act_range = tf.range(0, batch_size, dtype=tf.int64)[:, None]
    # Combine this into coordinates with the action indices
    indices = tf.concat([act_range, act_indx], 1)

    actions_mask = tf.SparseTensor(indices, values, [batch_size, num_actions])
    actions_mask = tf.stop_gradient(
        tf.sparse.to_dense(actions_mask, default_value=False)
    )
    sliced_emb = tf.boolean_mask(tensor=embeddings, mask=actions_mask)
    return sliced_emb


def categorical_dist_qlearning(
    atoms_tm1,
    logits_q_tm1,
    a_tm1,
    r_t,
    pcont_t,
    atoms_t,
    logits_q_t,
    name="CategoricalDistQLearning",
):
    """Implements Distributional Q-learning as TensorFlow ops.

  The function assumes categorical value distributions parameterized by logits.

  See "A Distributional Perspective on Reinforcement Learning" by Bellemare,
  Dabney and Munos. (https://arxiv.org/abs/1707.06887).

  Args:
    atoms_tm1: 1-D tensor containing atom values for first timestep,
      shape `[num_atoms]`.
    logits_q_tm1: Tensor holding logits for first timestep in a batch of
      transitions, shape `[B, num_actions, num_atoms]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    atoms_t: 1-D tensor containing atom values for second timestep,
      shape `[num_atoms]`.
    logits_q_t: Tensor holding logits for second timestep in a batch of
      transitions, shape `[B, num_actions, num_atoms]`.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: a tensor containing the values that `q_tm1` at actions
        `a_tm1` are regressed towards, shape `[B, num_atoms]`.

  Raises:
    ValueError: If the tensors do not have the correct rank or compatibility.
  """
    # Rank and compatibility checks.
    assertion_lists = [
        [logits_q_tm1, logits_q_t],
        [a_tm1, r_t, pcont_t],
        [atoms_tm1, atoms_t],
    ]
    base_ops.wrap_rank_shape_assert(assertion_lists, [3, 1, 1], name)

    # Categorical distributional Q-learning op.
    with tf.compat.v1.name_scope(
        name, values=[atoms_tm1, logits_q_tm1, a_tm1, r_t, pcont_t, atoms_t, logits_q_t]
    ):

        with tf.compat.v1.name_scope("target"):
            # Scale and shift time-t distribution atoms by discount and reward.
            target_z = r_t[:, None] + pcont_t[:, None] * atoms_t[None, :]

            # Convert logits to distribution, then find greedy action in state s_t.
            q_t_probs = tf.nn.softmax(logits_q_t)
            q_t_mean = tf.reduce_sum(input_tensor=q_t_probs * atoms_t, axis=2)
            pi_t = tf.argmax(input=q_t_mean, axis=1, output_type=tf.int32)

            # Compute distribution for greedy action.
            p_target_z = _slice_with_actions(q_t_probs, pi_t)

            # Project using the Cramer distance
            target = tf.stop_gradient(_l2_project(target_z, p_target_z, atoms_tm1))

        logit_qa_tm1 = _slice_with_actions(logits_q_tm1, a_tm1)

        loss = tf.nn.softmax_cross_entropy_with_logits(
            logits=logit_qa_tm1, labels=tf.stop_gradient(target)
        )

        return base_ops.LossOutput(loss, Extra(target))


def categorical_dist_double_qlearning(
    atoms_tm1,
    logits_q_tm1,
    a_tm1,
    r_t,
    pcont_t,
    atoms_t,
    logits_q_t,
    q_t_selector,
    name="CategoricalDistDoubleQLearning",
):
    """Implements Distributional Double Q-learning as TensorFlow ops.

  The function assumes categorical value distributions parameterized by logits,
  and combines distributional RL with double Q-learning.

  See "Rainbow: Combining Improvements in Deep Reinforcement Learning" by
  Hessel, Modayil, van Hasselt, Schaul et al.
  (https://arxiv.org/abs/1710.02298).

  Args:
    atoms_tm1: 1-D tensor containing atom values for first timestep,
      shape `[num_atoms]`.
    logits_q_tm1: Tensor holding logits for first timestep in a batch of
      transitions, shape `[B, num_actions, num_atoms]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    atoms_t: 1-D tensor containing atom values for second timestep,
      shape `[num_atoms]`.
    logits_q_t: Tensor holding logits for second timestep in a batch of
      transitions, shape `[B, num_actions, num_atoms]`.
    q_t_selector: Tensor holding another set of Q-values for second timestep
      in a batch of transitions, shape `[B, num_actions]`.
      These values are used for estimating the best action. In Double DQN they
      come from the online network.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: Tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`:  Tensor containing the values that `q_tm1` at actions
        `a_tm1` are regressed towards, shape `[B, num_atoms]` .

  Raises:
    ValueError: If the tensors do not have the correct rank or compatibility.
  """
    # Rank and compatibility checks.
    assertion_lists = [
        [logits_q_tm1, logits_q_t],
        [a_tm1, r_t, pcont_t],
        [atoms_tm1, atoms_t],
        [q_t_selector],
    ]
    base_ops.wrap_rank_shape_assert(assertion_lists, [3, 1, 1, 2], name)

    # Categorical distributional double Q-learning op.
    with tf.compat.v1.name_scope(
        name,
        values=[
            atoms_tm1,
            logits_q_tm1,
            a_tm1,
            r_t,
            pcont_t,
            atoms_t,
            logits_q_t,
            q_t_selector,
        ],
    ):

        with tf.compat.v1.name_scope("target"):
            # Scale and shift time-t distribution atoms by discount and reward.
            target_z = r_t[:, None] + pcont_t[:, None] * atoms_t[None, :]

            # Convert logits to distribution, then find greedy policy action in
            # state s_t.
            q_t_probs = tf.nn.softmax(logits_q_t)
            pi_t = tf.argmax(input=q_t_selector, axis=1, output_type=tf.int32)
            # Compute distribution for greedy action.
            p_target_z = _slice_with_actions(q_t_probs, pi_t)

            # Project using the Cramer distance
            target = tf.stop_gradient(_l2_project(target_z, p_target_z, atoms_tm1))

        logit_qa_tm1 = _slice_with_actions(logits_q_tm1, a_tm1)

        loss = tf.nn.softmax_cross_entropy_with_logits(
            logits=logit_qa_tm1, labels=tf.stop_gradient(target)
        )

        return base_ops.LossOutput(loss, Extra(target))


def categorical_dist_td_learning(
    atoms_tm1,
    logits_v_tm1,
    r_t,
    pcont_t,
    atoms_t,
    logits_v_t,
    name="CategoricalDistTDLearning",
):
    """Implements Distributional TD-learning as TensorFlow ops.

  The function assumes categorical value distributions parameterized by logits.

  See "A Distributional Perspective on Reinforcement Learning" by Bellemare,
  Dabney and Munos. (https://arxiv.org/abs/1707.06887).

  Args:
    atoms_tm1: 1-D tensor containing atom values for first timestep,
      shape `[num_atoms]`.
    logits_v_tm1: Tensor holding logits for first timestep in a batch of
      transitions, shape `[B, num_atoms]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    atoms_t: 1-D tensor containing atom values for second timestep,
      shape `[num_atoms]`.
    logits_v_t: Tensor holding logits for second timestep in a batch of
      transitions, shape `[B, num_atoms]`.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: Tensor containing the batch of losses, shape `[B]`.
    * `extra`: A namedtuple with fields:
        * `target`: Tensor containing the values that `v_tm1` are
        regressed towards, shape `[B, num_atoms]`.

  Raises:
    ValueError: If the tensors do not have the correct rank or compatibility.
  """
    # Rank and compatibility checks.
    assertion_lists = [[logits_v_tm1, logits_v_t], [r_t, pcont_t], [atoms_tm1, atoms_t]]
    base_ops.wrap_rank_shape_assert(assertion_lists, [2, 1, 1], name)

    # Categorical distributional TD-learning op.
    with tf.compat.v1.name_scope(
        name, values=[atoms_tm1, logits_v_tm1, r_t, pcont_t, atoms_t, logits_v_t]
    ):

        with tf.compat.v1.name_scope("target"):
            # Scale and shift time-t distribution atoms by discount and reward.
            target_z = r_t[:, None] + pcont_t[:, None] * atoms_t[None, :]
            v_t_probs = tf.nn.softmax(logits_v_t)

            # Project using the Cramer distance
            target = tf.stop_gradient(_l2_project(target_z, v_t_probs, atoms_tm1))

        loss = tf.nn.softmax_cross_entropy_with_logits(
            logits=logits_v_tm1, labels=tf.stop_gradient(target)
        )

        return base_ops.LossOutput(loss, Extra(target))




from typing import Tuple
from ..mathy.agents.episode_memory import EpisodeMemory
from ..mathy.envs.poly_simplify import PolySimplify
from numpy.testing import assert_array_almost_equal


def get_memory(number_observations: int) -> Tuple[EpisodeMemory, PolySimplify]:
    memory = EpisodeMemory()
    env = PolySimplify()
    state, problem = env.get_initial_state()
    for i in range(number_observations):
        memory.store(
            observation=env.state_to_observation(state),
            action=0,
            reward=0.0,
            value=0.0,
        )
    return memory, env


def test_episode_memory_to_episode_window():
    """verify that to_episode_window returns a window of the entire episode"""
    memory, env = get_memory(12)

    assert len(memory.observations) == 12

    assert len(memory.to_episode_window().nodes) == 12


def test_episode_memory_to_window_observation():
    """Verify that to_window_observation only returns a subset of the memory"""
    memory, env = get_memory(4)
    assert len(memory.observations) == 4
    obs = env.state_to_observation(env.get_initial_state()[0])
    # Returns only the last three
    assert len(memory.to_window_observation(obs, window_size=3).nodes) == 3
    # Only returns as many observations as it has (plus the one provided)
    assert len(memory.to_window_observation(obs, window_size=10).nodes) == 5
